/*
 * yasapi_player.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd316756%28v=vs.85%29.aspx
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd370844%28v=vs.85%29.aspx

///////////////////////////////////////////////////////////////////////////////
#define PLAYER_FRAMES_MS(nFrames,pwfx) \
  MulDiv(nFrames,1000,(pwfx)->nSamplesPerSec)
#define PLAYER_FRAMES_TIMER(nFrames,pwfx) \
  (10000000.0*(nFrames)/(pwfx)->nSamplesPerSec)

///////////////////////////////////////////////////////////////////////////////
int PlayerCreate(Player *pPlayer, Plugin *pPlugin, int samplerate,
    int numchannels, int bitspersamp)
{
  Options *pOptions=pPlugin->pOptions;
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  AUDCLNT_SHAREMODE eShareMode=pOptions->eShareMode;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  IMMDeviceEnumerator *pEnumerator=pPlugin->pEnumerator;
  IMMDevice *pDevice=pPlugin->pDevice;
  HRESULT hr;
  UINT32 nBlockAlign;
  IAudioClient *pAudioClient;
  IAudioRenderClient *pRenderClient;
  IAudioClock *pAudioClock;
#if defined (YASAPI_DEBUG) // {
  WAVEFORMATEX *pDeviceFormat;
#endif // }
  WAVEFORMATEX *pClosestMatch,**ppClosestMatch;
  REFERENCE_TIME hnsDevicePeriod;
  UINT32 nBufSize;
  SyncConfig sc;

  if (0==pPlugin->initialized) {
    DMESSAGE(debug,"not initialized");
    goto initialized;
  }

  pPlayer->strategy=pOptions->pull?PlayerPull():PlayerPush();
  DPRINTF(debug,"  initializing for %s mode\n",pPlayer->strategy->name);
  pPlayer->pPlugin=pPlugin;

  // setup WAVEFORMATEXTENSIBLE.
  if (1==numchannels&&0!=pOptions->mono2stereo) {
    pPlayer->mono=1;
    numchannels=2;
  }
  else
    pPlayer->mono=0;

  WFXXSetup(&pPlayer->wfxx,samplerate,numchannels,bitspersamp);
  pPlayer->wBytesPerSample=pwfx->wBitsPerSample>>3;

  // from the device, get an audio client.
  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,        // [in]   REFIID iid,
    CLSCTX_ALL,               // [in]   DWORD dwClsCtx,
    NULL,                     // [in]   PROPVARIANT *pActivationParams,
    (void **)&pAudioClient    // [out]  void **ppInterface
  );

  if (FAILED(hr)) {
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DMESSAGE(debug,"activating audio client");
    goto client;
  }

  pPlayer->pAudioClient=pAudioClient;
  DPUTS(debug,"  audio client activated\n");

#if defined (YASAPI_DEBUG) // {
  if (0!=debug) {
    hr=pAudioClient->lpVtbl->GetMixFormat(pAudioClient,
      &pDeviceFormat          // [out]  WAVEFORMATEX **ppDeviceFormat
    );

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DERROR(debug,E_POINTER,hr);
      DERROR(debug,E_OUTOFMEMORY,hr);
      DMESSAGE(debug,"getting device format");
      goto format;
    }
    else if (NULL!=pDeviceFormat) {
#if defined (YASAPI_DEBUG) // {
      if (debug) {
        fputs("  device format: ",stderr);
        WFXPrint(pDeviceFormat,stderr);
        fputs("\n",stderr);
      }
#endif // }
      CoTaskMemFree(pDeviceFormat);
    }
  }
#endif // }

  // check whether the format is supported.
  pClosestMatch=NULL;
  ppClosestMatch=AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode?NULL:&pClosestMatch;

  hr=pAudioClient->lpVtbl->IsFormatSupported(pAudioClient,
    eShareMode,             // [in]   AUDCLNT_SHAREMODE ShareMode,
    pwfx,                   // [in]   const WAVEFORMATEX *pFormat,
    ppClosestMatch          // [out]  WAVEFORMATEX **ppClosestMatch
  );

#if ! defined (YASAPI_RESAMPLE) // {
  if (FAILED(hr)||S_FALSE==hr) {
    DERROR(debug,S_FALSE,hr);
    DERROR(debug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGEV(debug,"format not supported (channels: %d, rate: %d, bps: %d)",
        pwfx->nChannels,pwfx->nSamplesPerSec,pwfx->wBitsPerSample);

    if (NULL!=pClosestMatch) {
#if defined (YASAPI_DEBUG) // {
      if (debug) {
        fputs("  closest match: ",stderr);
        WFXPrint(pClosestMatch,stderr);
        fputs("\n",stderr);
      }
#endif // }
      CoTaskMemFree(pClosestMatch);
    }

    goto format;
  }

#if defined (YASAPI_DEBUG) // {
    if (debug) {
      fputs("  format supported: ",stderr);
      WFXPrint(&pPlayer->wfxx.Format,stderr);
      fputs("\n",stderr);
    }
#endif // }
#else // } {
  if (S_OK==hr) {
#if defined (YASAPI_DEBUG) // {
    if (debug) {
      fputs("  format supported: ",stderr);
      WFXPrint(&pPlayer->wfxx.Format,stderr);
      fputs("\n",stderr);
    }
#endif // }
    pPlayer->write=PlayerWriteDirect();
    
    if (pPlayer->write->Create(pPlayer,NULL)<0) {
      DMESSAGEV(debug,"creating %s write startegy",pPlayer->write->name);
      goto format;
    }
  }
  else if (S_FALSE==hr) {
#if defined (YASAPI_DEBUG) // {
    if (debug) {
      fputs("  closest match supported: ",stderr);
      WFXPrint(pClosestMatch,stderr);
      fputs("\n",stderr);
    }
#endif // }
    pPlayer->write=PlayerWriteIndirect();
    
    if (pPlayer->write->Create(pPlayer,pClosestMatch)<0) {
      DMESSAGEV(debug,"creating %s write startegy",pPlayer->write->name);
      goto format;
    }
  }
  else {
    //DERROR(debug,S_FALSE,hr);
    DERROR(debug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGEV(debug,"format not supported (channels: %d, rate: %d, bps: %d)",
        pwfx->nChannels,pwfx->nSamplesPerSec,pwfx->wBitsPerSample);
    goto format;
  }
#endif // }

  // get the device period.
#if 1 // {
  // get the minimum device period (low latency).
  hr=pAudioClient->lpVtbl->GetDevicePeriod(pAudioClient,
    NULL,                 // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
    &hnsDevicePeriod      // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
  );
#else // } {
  // get the default device period.
  hr=pAudioClient->lpVtbl->GetDevicePeriod(pAudioClient,
    &hnsDevicePeriod,     // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
    NULL                  // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
  );
#endif // }

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting device period");
    goto period;
  }

  DPRINTF(debug,"  got device period (device period: %d)\n",hnsDevicePeriod);

  if (1.0<pOptions->qShareSize) {
    hnsDevicePeriod=pOptions->qShareSize*hnsDevicePeriod+0.5;
    DPRINTF(debug,"  device period scaled by %.02f (device period: %d)\n",
        pOptions->qShareSize,hnsDevicePeriod);
  }

  hr=pPlayer->strategy->InitializeAudioClient(pPlayer,hnsDevicePeriod);

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_ALREADY_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_CPUUSAGE_EXCEEDED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_IN_USE,hr);
    DERROR(debug,AUDCLNT_E_ENDPOINT_CREATE_FAILED,hr);
    DERROR(debug,AUDCLNT_E_INVALID_DEVICE_PERIOD,hr);
    DERROR(debug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(debug,AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED,hr);
    DERROR(debug,AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGEV(debug,"initializing audio client (mode: %s)",
        YASAPI_SHAREMODE_LABEL(eShareMode));
    goto initialize;
  }

  DPRINTF(debug,"  audio client initialized (mode: %s)\n",
      YASAPI_SHAREMODE_LABEL(eShareMode));

  if (pPlayer->strategy->CreateEvent(pPlayer)<0)
    goto event;

  // get the buffer size.
  //
  // This method retrieves the length of the endpoint buffer shared
  // between the client application and the audio engine. The length
  // is expressed as the number of audio frames the buffer can hold.
  // The size in bytes of an audio frame is calculated as the number
  // of channels in the stream multiplied by the sample size per channel.
  // For example, the frame size is four bytes for a stereo (2-channel)
  // stream with 16-bit samples.
  hr=pAudioClient->lpVtbl->GetBufferSize(pAudioClient,
    &pPlayer->nFrames         // [out]  UINT32 *pNumBufferFrames
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting buffer size");
    goto size;
  }

  DPRINTF(debug,"  got buffer size (frames: %d)\n",pPlayer->nFrames);

  // get the render client.
  hr=pAudioClient->lpVtbl->GetService(pAudioClient,
    &IID_IAudioRenderClient,  // [in]   REFIID riid,
    (void**)&pRenderClient    // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting render audio client");
    goto render;
  }

  pPlayer->pRenderClient=pRenderClient;
  DPUTS(debug,"  got render audio client\n");

  // create the sync.
#if defined (YASAPI_DEBUG) // {
  sc.debug=debug;
#endif // }
  sc.pClient=pPlayer;
  sc.end.state=PLAYER_END;
  sc.read.state=PLAYER_READ;
  sc.read.Condition=PlayerCanRead;
  sc.write.state=PLAYER_WRITE;
  sc.write.Condition=PlayerCanWrite;

  if (SyncCreate(&pPlayer->sync,&sc)<0) {
    DMESSAGE(debug,"creating sync");
    goto sync;
  }

  DPUTS(debug,"  sync created\n");

  // create the ring buffer.
  if (1.0<pOptions->qRingSize) {
    nBufSize=pOptions->qRingSize*pPlayer->nFrames*pwfx->nBlockAlign+0.5;

    if (0<nBufSize%pwfx->nBlockAlign) {
      nBufSize+=pwfx->nBlockAlign;
      nBufSize-=nBufSize%pwfx->nBlockAlign;
    }
  }
  else
    nBufSize=pPlayer->nFrames*pwfx->nBlockAlign;

  if (RingCreate(&pPlayer->ring,nBufSize,debug)<0) {
    DMESSAGEV(debug,"creating ring buffer (scale %.1f, size: %d)",
        pOptions->qRingSize,nBufSize);
    goto ring;
  }

  pPlayer->ring.Copy=PlayerCopyMemory;
  pPlayer->ring.pClient=pPlayer;

  DPRINTF(debug,"  ring buffer created (scale %.1f, size: %d)\n",
      (double)nBufSize/(pwfx->nBlockAlign*pPlayer->nFrames)+0.5,nBufSize);

  // create the render audio thread.
  pPlayer->hThread=CreateThread(
    NULL,           // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,              // _In_       SIZE_T dwStackSize,
    PlayerThread,   // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    pPlayer,        // _In_opt_   LPVOID lpParameter,
    0,              // _In_       DWORD dwCreationFlags,
    NULL            // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==pPlayer->hThread) {
    DMESSAGE(debug,"creating render audio thread");
    goto thread;
  }

  DPUTS(debug,"  render audio thread created\n");

  // get the render client.
  hr=pAudioClient->lpVtbl->GetService(pAudioClient,
    &IID_IAudioClock,  // [in]   REFIID riid,
    (void**)&pAudioClock    // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting audio clock");
    goto clock;
  }

  pPlayer->pAudioClock=pAudioClock;
  DPUTS(debug,"  got audio clock\n");

  return 0;
// cleanup:
  pPlayer->pAudioClock=NULL;
  EnterCriticalSection(&pPlugin->cs);
  pAudioClock->lpVtbl->Release(pAudioClock);
  LeaveCriticalSection(&pPlugin->cs);
clock:
thread:
  RingDestroy(&pPlayer->ring);
ring:
  SyncDestroy(&pPlayer->sync);
sync:
  pRenderClient->lpVtbl->Release(pRenderClient);
render:
size:
  pPlayer->strategy->DestroyEvent(pPlayer);
event:
initialize:
period:
#if defined (YASAPI_RESAMPLE) // {
  pPlayer->write->Destroy(pPlayer);
#endif // }
format:
  pAudioClient->lpVtbl->Release(pAudioClient);
client:
initialized:
  return -1;
}

void PlayerDestroy(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Plugin *pPlugin=pPlayer->pPlugin;
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  IAudioClock *pAudioClock=pPlayer->pAudioClock;
  IAudioRenderClient *pRenderClient=pPlayer->pRenderClient;

  pPlayer->pAudioClock=NULL;
  EnterCriticalSection(&pPlugin->cs);
  pAudioClock->lpVtbl->Release(pAudioClock);
  LeaveCriticalSection(&pPlugin->cs);
  DPUTS(debug,"  render audio clock\n");

  WaitForSingleObject(pSync->hMutex,INFINITE);
  pSync->state|=PLAYER_END;
  SetEvent(pSync->read.hEvent);
  SetEvent(pSync->write.hEvent);
  ReleaseMutex(pSync->hMutex);

  WaitForSingleObject(pPlayer->hThread,INFINITE);
  CloseHandle(pPlayer->hThread);
  DPUTS(debug,"  render audio thread destroyed\n");

  RingDestroy(&pPlayer->ring);
  DPUTS(debug,"  ring buffer destroyed\n");
  SyncDestroy(&pPlayer->sync);
  DPUTS(debug,"  sync destroyed\n");

  pPlayer->strategy->DestroyEvent(pPlayer);

  pRenderClient->lpVtbl->Release(pRenderClient);
  DPUTS(debug,"  render audio client destroyed\n");
#if defined (YASAPI_RESAMPLE) // {
  pPlayer->write->Destroy(pPlayer);
#endif // }
  pAudioClient->lpVtbl->Release(pAudioClient);
  DPUTS(debug,"  audio client destroyed\n");

  pPlayer->pPlugin=NULL;
}

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_DEBUG) // {
int PlayerGetDebug(Player *pPlayer)
{
  return pPlayer->pPlugin->pOptions->debug;
}
#endif // }

HRESULT PlayerStart(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr=S_OK;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  if (0==((PLAYER_STARTED|PLAYER_END)&pSync->state)) {
    // start the audio client.
    hr=pAudioClient->lpVtbl->Start(pAudioClient);

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
      DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_SET,hr);
      DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DMESSAGE(debug,"starting audio client");
      goto start;
    }

    pSync->state|=PLAYER_STARTED;
    DPUTS(debug,"  audio client started\n");
  }
start:
  ReleaseMutex(pSync->hMutex);
  return hr;
}

HRESULT PlayerStopAtom(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr=S_OK;

  if (PLAYER_IS_STARTED(pSync)) {
    // start the audio client.
    hr=pAudioClient->lpVtbl->Stop(pAudioClient);

    if (FAILED(hr)) {
      DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
      DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING ,hr);
      DMESSAGE(debug,"stopping audio client");
      goto start;
    }

    DPUTS(debug,"  audio client stopped\n");
    pSync->state&=~PLAYER_STARTED;
  }
start:
  return hr;
}

HRESULT PlayerStop(Player *pPlayer)
{
  HANDLE hMutex=pPlayer->sync.hMutex;
  HRESULT hr;

  WaitForSingleObject(hMutex,INFINITE);
  hr=PlayerStopAtom(pPlayer);
  ReleaseMutex(hMutex);

  return hr;
}

DWORD PlayerGetTime(Player *pPlayer, Plugin *pPlugin)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  DWORD dwTime=0;
  IAudioClock *pAudioClock;
  UINT64 u64Position;
  UINT64 u64Frequency;
  HRESULT hr;

  EnterCriticalSection(&pPlugin->cs);
  pAudioClock=pPlayer->pAudioClock;

  // not initialized.
  if (NULL==pAudioClock)
    goto clock;

  // getting the position.
  hr=pAudioClock->lpVtbl->GetPosition(pAudioClock,
    &u64Position,     // [out]  UINT64 *pu64Position,
    NULL              // [out]  UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    DERROR(debug,S_FALSE,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting position");
    goto position;
  }

  // getting the freqency.
  hr=pAudioClock->lpVtbl->GetFrequency(pAudioClock,
    &u64Frequency     // [out]  UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"getting frequency");
    goto frequency;
  }

  dwTime=1000.0*u64Position/u64Frequency+0.5;
frequency:
position:
clock:
  LeaveCriticalSection(&pPlugin->cs);
  return dwTime;
}

WORD PlayerSamplesPerSec(Player *pPlayer)
{
  return pPlayer->wfxx.Format.nSamplesPerSec;
}

WORD PlayerChannels(Player *pPlayer)
{
  return 0==pPlayer->mono?pPlayer->wfxx.Format.nChannels:1;
}

WORD PlayerBlockAlign(Player *pPlayer)
{
  return PlayerChannels(pPlayer)*pPlayer->wBytesPerSample;
}

SIZE_T PlayerAvailable(Player *pPlayer)
{
  return 0==pPlayer->mono
      ?pPlayer->ring.dwAvailable
      :pPlayer->ring.dwAvailable>>1;
}

BOOL PlayerCanWrite(LPVOID pClient)
{
  return 0<((Player *)pClient)->ring.dwAvailable;
}

BOOL PlayerCanRead(LPVOID pClient)
{
  Player *pPlayer=pClient;
  Sync *pSync=&pPlayer->sync;

  return !PLAYER_IS_PAUSE(pSync)&&pPlayer->strategy->CanRead(pPlayer);
}

////////
SIZE_T PlayerWriteWait(Player *pPlayer, LPSTR pData, SIZE_T dwSize)
{
#if defined (YASAPI_DEBUG) // {
  int debug=PlayerGetDebug(pPlayer);
#endif // }
  Sync *pSync=&pPlayer->sync;
  Ring *pRing=&pPlayer->ring;

  dwSize-=dwSize%pPlayer->wfxx.Format.nBlockAlign;

  if (0!=pPlayer->mono)
    dwSize<<=1;

  SyncWaitWrite(pSync);
//DPRINTF(debug,"WRITE (%d, %d) -> ",pRing->dwWritten,pRing->dwAvailable);
  dwSize=RingWrite(pRing,pData,dwSize,0==pPlayer->mono?1:2);
//DPRINTF(debug,"(%d, %d)\n",pRing->dwWritten,pRing->dwAvailable);
  SyncReleaseWrite(pSync);

  if (0!=pPlayer->mono)
    dwSize>>=1;

  return dwSize;
}

SIZE_T PlayerWrite(Player *pPlayer, LPSTR pData, SIZE_T dwSize)
{
#if defined (YASAPI_RESAMPLE) // {
  return pPlayer->write->Write(pPlayer,pData,dwSize);
#else // } {
  return PlayerWriteWait(pPlayer,pData,dwSize);
#endif // }
}

HRESULT PlayerResetAtom(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  // get reset.
  hr=pAudioClient->lpVtbl->Reset(pAudioClient);

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DMESSAGE(debug,"resetting audio client");
    goto reset;
  }
reset:
  return hr;
}

void PlayerReset(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  // stop the player.
  if (PLAYER_IS_STARTED(pSync)) {
    hr=PlayerStopAtom(pPlayer);

    if (FAILED(hr)) {
      DMESSAGE(debug,"stopping audio client");
      goto stop;
    }
  }

  hr=PlayerResetAtom(pPlayer);

  if (FAILED(hr)) {
    DMESSAGE(debug,"resetting audio client");
    goto reset;
  }

  // reset the ring buffer.
  RingReset(&pPlayer->ring);
reset:
stop:
  ReleaseMutex(pSync->hMutex);
}

void PlayerPause(Player *pPlayer, BOOL bPause)
{
  Sync *pSync=&pPlayer->sync;

  WaitForSingleObject(pSync->hMutex,INFINITE);

  if (0==bPause)
    pSync->state&=~PLAYER_PAUSE;
  else {
    pSync->state|=PLAYER_PAUSE;
    PlayerStopAtom(pPlayer);
  }

  SetEvent(pSync->read.hEvent);
  SetEvent(pSync->write.hEvent);
  ReleaseMutex(pSync->hMutex);
}

///////////////////////////////////////////////////////////////////////////////
HRESULT PlayerRenderFramesUnlocked(Player *pPlayer, UINT32 nFramesWrite,
    DWORD dwFlags)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  IAudioRenderClient *pRenderClient=pPlayer->pRenderClient;
  Ring *pRing=&pPlayer->ring;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  BYTE *pData;
  HRESULT hr;

  hr=pRenderClient->lpVtbl->GetBuffer(pRenderClient,
    nFramesWrite,       // [in]   UINT32 NumFramesRequested,
    &pData              // [out]  BYTE **ppData
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_BUFFER_ERROR,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_TOO_LARGE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting render buffer");
    goto get;
  };


  if (0<nFramesWrite) {
    // transfer the audio data from the ring buffer into the render buffer.
//DPRINTF(debug,"READ (%d, %d) -> ",pRing->dwWritten,pRing->dwAvailable);
    RingRead(pRing,pData,nFramesWrite*pwfx->nBlockAlign);
//DPRINTF(debug,"(%d, %d)\n",pRing->dwWritten,pRing->dwAvailable);
  }

  // release the render buffer.
  hr=pRenderClient->lpVtbl->ReleaseBuffer(pRenderClient,
    nFramesWrite,     // [in]  UINT32 NumFramesWritten,
    dwFlags           // [in]  DWORD dwFlags
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_INVALID_SIZE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DMESSAGE(debug,"releasing render buffer");
    goto release;
  }
release:
get:
  return hr;
}

HRESULT PlayerRenderFrames(Player *pPlayer, UINT32 *pNumFramesRender)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#endif // }
  Sync *pSync=&pPlayer->sync;
  UINT32 nFramesPadding;
  UINT32 nFramesWrite;
  HRESULT hr;

  // lock the ring buffer.
  SyncWaitRead(pSync);

  if (PLAYER_IS_PAUSE(pSync)) {
    // if we are going here and are in "pause" we're in "end" too.
    if (!PLAYER_IS_END(pSync))
      DMESSAGE(debug,"inconsistent state");

    RingReset(&pPlayer->ring);
    *pNumFramesRender=0;
  }
  else {
    hr=pPlayer->strategy->GetFramesWrite(pPlayer,&nFramesWrite,
        &nFramesPadding);

    if (FAILED(hr))
      goto frames;

    // get the render buffer.
    if (0<nFramesWrite) {
      hr=PlayerRenderFramesUnlocked(pPlayer,nFramesWrite,0);

      if (FAILED(hr)) {
        DMESSAGE(debug,"rendering");
        goto render;
      }
    }

    *pNumFramesRender=nFramesPadding+nFramesWrite;
  }
render:
frames:
  // release the lock on the ring buffer.
  SyncReleaseRead(pSync);
  return hr;
}

void PlayerFlush(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Sync *pSync=&pPlayer->sync;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  UINT32 nFrames,nFramesRender;
  HRESULT hr;

  PlayerRenderFramesUnlocked(pPlayer,0,AUDCLNT_BUFFERFLAGS_SILENT);

  if (PLAYER_IS_PAUSE(pSync))
    PlayerResetAtom(pPlayer);
  else {
    nFrames=0;

    for (;;) {
      // get number of remaining frames in render buffer.
      hr=pAudioClient->lpVtbl->GetCurrentPadding(pAudioClient,
        &nFramesRender      // [out]  UINT32 *pNumPaddingFrames
      );

      if (FAILED(hr))
        break;
      else if (nFrames==nFramesRender) {
#if defined (YASAPI_DEBUG) // {
        if (0<nFrames)
          DPRINTF(debug, "  sorry, %d frames lost ...\n",nFrames);
#endif // }
        break;
      }

      // remember the numer of frames in order to avoid an endless loop.
      nFrames=nFramesRender;

      // in case there are frames, wait until played.
      TimerSleep(&pPlugin->timer,PLAYER_FRAMES_TIMER(nFrames,pwfx)+0.5);
    }
  }
}

DWORD WINAPI PlayerThread(LPVOID lpParameter)
{
  Player *pPlayer=lpParameter;
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Options *pOptions=pPlugin->pOptions;
  Sync *pSync=&pPlayer->sync;
  Ring *pRing=&pPlayer->ring;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  UINT32 nFrames=pPlayer->nFrames;
  LONGLONG time=0.5*PLAYER_FRAMES_TIMER(nFrames,pwfx)+0.5;
  UINT32 nFramesRender;
  HANDLE hTask;
  DWORD taskIndex;
  HRESULT hr;

  DPUTS(debug,"start playing\n");
  taskIndex=0;
  hTask=AvSetMmThreadCharacteristicsA("Pro Audio",&taskIndex);

  if (NULL==hTask) {
    DMESSAGE(debug,"setting thread cahracteristics");
    goto task;
  }

  if (1.0<pOptions->qRingFill)
    nFrames=pOptions->qRingFill*nFrames+0.5;

  while (!PLAYER_IS_END(pSync)) {
    if (PLAYER_IS_STARTED(pSync))
      pPlayer->strategy->Wait(pPlayer,time);

    hr=PlayerRenderFrames(pPlayer,&nFramesRender);

    if (FAILED(hr)) {
      DMESSAGE(debug,"rendering frames");
      continue;
    }

    // make sure that the renderer can be completely filled before starting.
    if (!PLAYER_IS_STARTED(pSync)&&nFrames<=pRing->dwWritten)
      PlayerStart(pPlayer);
  }

  if (PLAYER_IS_STARTED(pSync)) {
    PlayerFlush(pPlayer);
    PlayerStop(pPlayer);
  }

  DPUTS(debug,"  stop playing\n");
// cleanup:
  AvRevertMmThreadCharacteristics(hTask);
task:
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
void PlayerCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length)
{
  Player *pPlayer=pClient;
  Plugin *pPlugin=pPlayer->pPlugin;
  Options *pOptions=pPlugin->pOptions;
  double volume=pPlugin->volume;
  char *wp,*mp;
  const char *rp;
  int m,n;
  double q;
  int32_t i32;

  wp=Destination;
  mp=wp+Length;
  rp=Source;
  n=pPlayer->wfxx.Format.wBitsPerSample;
  m=n>>3;
  q=volume*YASAPI_INT_MAX(n)/YASAPI_INT_MAX(32);

  n=32-n;

  if (0==pPlayer->mono) {
    if (0==pOptions->volume||1.0==volume) {
      CopyMemory(
        Destination,            // _In_  PVOID Destination,
        Source,                 // _In_  const VOID *Source,
        Length                  // _In_  SIZE_T Length
      );
    }
    else {
      while (wp<mp) {
        CopyMemory(
          &i32,                 // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        i32<<=n;
        i32=q*i32+0.5;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
    }
  }
  else {
    while (wp<mp) {
      CopyMemory(
        wp,                     // _In_  PVOID Destination,
        rp,                     // _In_  const VOID *Source,
        m                       // _In_  SIZE_T Length
      );

      wp+=m;

      CopyMemory(
        wp,                     // _In_  PVOID Destination,
        rp,                     // _In_  const VOID *Source,
        m                       // _In_  SIZE_T Length
      );

      wp+=m;
      rp+=m;
    }
  }
}
