/*
 * yasapi_plugin.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

#if defined (YASAPI_RESAMPLE) // {
///////////////////////////////////////////////////////////////////////////////
// https://bug825153.bugzilla.mozilla.org/attachment.cgi?id=729339
#define MF_VERSION_VISTA    (0x0001<<16|MF_API_VERSION)
#define MF_VERSION_WIN7     (0x0002<<16|MF_API_VERSION)
#define OS_IS_VISTA(posv)   (6==(posv)->dwMajorVersion)
#endif // }

///////////////////////////////////////////////////////////////////////////////
int PluginCreate(Plugin *pPlugin, Options *pOptions)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#else // } {
  int debug=0;
#endif // }
#if defined (YASAPI_RESAMPLE) // {
  OSVERSIONINFO osv;
#endif // }
  IMMDeviceEnumerator *pEnumerator;
  IMMDevice *pDevice;
  HRESULT hr;

#if defined (YASAPI_RESAMPLE) // {
  // get os version.
  ZeroMemory(
    &osv,                       // [in]  PVOID Destination,
    sizeof osv                  // [in]  SIZE_T Length
  );

  osv.dwOSVersionInfoSize=sizeof osv;

  GetVersionEx(
    &osv                        // _Inout_  LPOSVERSIONINFO lpVersionInfo
  );

  DPRINTF(debug,"  os: %d.%d (build %d)\n",osv.dwMajorVersion,
      osv.dwMinorVersion,osv.dwBuildNumber);

  // starting up media foundation.
  hr=MFStartup(
    OS_IS_VISTA(&osv)?MF_VERSION_VISTA:MF_VERSION_WIN7,
                                // ULONG Version,
    MFSTARTUP_NOSOCKET          // DWORD dwFlags = MFSTARTUP_FULL
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_BAD_STARTUP_VERSION,hr);
    DERROR(debug,MF_E_DISABLED_IN_SAFEMODE,hr);
    DERROR(debug,E_NOTIMPL,hr);
    DMESSAGE(debug,"starting up media foundation");
    goto mfstartup;
  }

  DPUTS(debug,"  media foundation started up\n");
#endif // }

  // create the critical section.
  InitializeCriticalSection(&pPlugin->cs);
  DPUTS(debug,"  critical section created\n");

  pPlugin->pOptions=pOptions;
  pPlugin->volume=1.0;

  // create the timer.
  if (TimerCreate(&pPlugin->timer,debug)<0) {
    DMESSAGE(debug,"creating timer");
    goto timer;
  }

  DPUTS(debug,"  waitable timer created\n");

  // create a device enumerator.
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,  // _In_   REFCLSID rclsid,
    NULL,                       // _In_   LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                 // _In_   DWORD dwClsContext,
    &IID_IMMDeviceEnumerator,   // _In_   REFIID riid,
    (void**)&pEnumerator        // _Out_  LPVOID *ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,REGDB_E_CLASSNOTREG,hr);
    DERROR(debug,CLASS_E_NOAGGREGATION,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating device enumerator");
    goto devnum;
  }

  pPlugin->pEnumerator=pEnumerator;
  DPUTS(debug,"  device enumerator created\n");

  // get the default audio device.
  hr=pEnumerator->lpVtbl->GetDefaultAudioEndpoint(pEnumerator,
    eRender,                  // [in]   EDataFlow dataFlow,
    eMultimedia,              // [in]   ERole role,
    &pDevice                  // [out]  IMMDevice **ppDevice
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_NOTFOUND,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"getting default audio device");
    goto device;
  }

  pPlugin->pDevice=pDevice;
  DPUTS(debug,"  device created\n");

  // yep!
  pPlugin->initialized=1;

  return 0;
// cleanup:
  pDevice->lpVtbl->Release(pDevice);
device:
  pEnumerator->lpVtbl->Release(pEnumerator);
devnum:
  TimerDestroy(&pPlugin->timer);
timer:
#if defined (YASAPI_RESAMPLE) // {
  MFShutdown();
mfstartup:
#endif // }
  DeleteCriticalSection(&pPlugin->cs);
  return -1;
}

void PluginDestroy(Plugin *pPlugin)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->pOptions->debug;
#endif // }
  IMMDeviceEnumerator *pEnumerator=pPlugin->pEnumerator;
  IMMDevice *pDevice=pPlugin->pDevice;

  pPlugin->initialized=0;

  pDevice->lpVtbl->Release(pDevice);
  DPUTS(debug,"  device destroyed\n");
  pEnumerator->lpVtbl->Release(pEnumerator);
  DPUTS(debug,"  device enumerator destroyed\n");
  TimerDestroy(&pPlugin->timer);
  DPUTS(debug,"  waitable timer destroyed\n");

  DeleteCriticalSection(&pPlugin->cs);
  DPUTS(debug,"  critical section destroyed\n");

#if defined (YASAPI_RESAMPLE) // {
  MFShutdown();
  DPUTS(debug,"  media foundation shut down\n");
#endif // }
}
