/*
 * yasapi_resample.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <d3d9.h>

///////////////////////////////////////////////////////////////////////////////
// https://code.google.com/p/bitspersampleconv2/wiki/HowToUseResamplerMFT

///////////////////////////////////////////////////////////////////////////////
#define BUF_SIZE 8192

#if 0 // {
///////////////////////////////////////////////////////////////////////////////
// https://bug825153.bugzilla.mozilla.org/attachment.cgi?id=729339
#define MF_VERSION_VISTA    (0x0001<<16|MF_API_VERSION)
#define MF_VERSION_WIN7     (0x0002<<16|MF_API_VERSION)
#endif // }

#if defined (YASAPI_RESAMPLE) // {
/*
 * We've to create several objects:
 *   * The resampler itself is an IMFTransform.
 *     # It's created indirectly via an IUnknown.
 *     # For configuration it needs
 *       - an IWMResamplerProps,
 *       - an IMFMediaType for input, and
 *       - an IMFMediaType for output.
 *   * An IMFSample, a collection of IMFMediaBuffers.
 *   * An IMFMediaBuffer.
 */

///////////////////////////////////////////////////////////////////////////////
int ResampleCreate(Player *pPlayer, WAVEFORMATEX *pwfx)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  Resample *pResample=&pPlayer->res;
  IUnknown *pTransformUnk;
  IMFTransform *pTransform;
  IWMResamplerProps *pResamplerProps;
  IMFSample *pSample;
  IMFMediaBuffer *pBuffer;
  HRESULT hr;

  ZeroMemory(
    pResample,                      // [in]  PVOID Destination,
    sizeof *pResample               // [in]  SIZE_T Length
  );

  pResample->pPlayer=pPlayer;
  pResample->pwfx=pwfx;

  // create a transform unknown.
  hr=CoCreateInstance(
    &CLSID_CResamplerMediaObject,   // _In_   REFCLSID rclsid,
    NULL,                           // _In_   LPUNKNOWN pUnkOuter,
    CLSCTX_INPROC_SERVER,           // _In_   DWORD dwClsContext,
    &IID_IUnknown,                  // _In_   REFIID riid,
    (LPVOID *)&pTransformUnk        // _Out_  LPVOID *ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,REGDB_E_CLASSNOTREG,hr);
    DERROR(debug,CLASS_E_NOAGGREGATION,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating transform unknown");
    goto factory;
  }

  pResample->pTransformUnk=pTransformUnk;
  DPUTS(debug,"  transform unknown created\n");

  // create the transform.
  hr=pTransformUnk->lpVtbl->QueryInterface(pTransformUnk,
    &IID_IMFTransform,            // [in]   REFIID riid,
    (void **)&pTransform          // [out]  void **ppvObject
  );

  if (FAILED(hr)) {
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating transform");
    goto transform;
  }

  pResample->pTransform=pTransform;
DPRINTF(debug,"0.1: %p ...\n",pTransform);
  DPUTS(debug,"  transform created\n");

  // create resample properties.
  hr=pTransformUnk->lpVtbl->QueryInterface(pTransformUnk,
    &IID_IWMResamplerProps,       // [in]   REFIID riid,
    (void **)&pResamplerProps     // [out]  void **ppvObject
  );

  if (FAILED(hr)) {
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating resample properties");
    goto props;
  }

  pResample->pResamplerProps=pResamplerProps;
  DPUTS(debug,"  resample properties created\n");

  // best quality, always success.
  hr=pResamplerProps->lpVtbl->SetHalfFilterLength(pResamplerProps,
    60                            // [in]  LONG lhalfFilterLen
  );

  // create the input channel.
  if (ResampleChannelCreate(&pResample->input,&pPlayer->wfxx.Format,
      pTransform->lpVtbl->SetInputType,pTransform,"input",debug)<0)
    goto input;

  pSample=pResample->input.pSample;

  // create the input media buffer.
  hr=MFCreateMemoryBuffer(
    BUF_SIZE,                   // DWORD cbMaxLength,
    &pBuffer                    // IMFMediaBuffer **ppBuffer
  );
  
  if (FAILED(hr)) {
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"creating input media buffer");
    goto buffer;
  }

  pResample->pBuffer=pBuffer;
  DPUTS(debug,"  memory input buffer created\n");

  // add the input media buffer to the input sample.
  hr=pSample->lpVtbl->AddBuffer(pSample,
    pBuffer                     // [in]  IMFMediaBuffer *pBuffer
  );

  if (FAILED(hr)) {
    DERROR(debug,E_INVALIDARG,hr);
    DMESSAGE(debug,"creating input media buffer");
    goto add;
  }

  // create the output channel.
  if (ResampleChannelCreate(&pResample->output,pResample->pwfx,
      pTransform->lpVtbl->SetOutputType,pTransform,"output",debug)<0)
    goto output;

  /////////////////////////////////////////////////////////////////////////////
  // process MFT_MESSAGE_COMMAND_FLUSH message.
  hr=TransformProcessMessage(pTransform,MFT_MESSAGE_COMMAND_FLUSH,
      "MFT_MESSAGE_COMMAND_FLUSH",debug);

  if (FAILED(hr))
    goto message;

  // process MFT_MESSAGE_NOTIFY_BEGIN_STREAMING message.
  hr=TransformProcessMessage(pTransform,MFT_MESSAGE_NOTIFY_BEGIN_STREAMING,
      "MFT_MESSAGE_NOTIFY_BEGIN_STREAMING",debug);

  if (FAILED(hr))
    goto message;

  // process MFT_MESSAGE_NOTIFY_START_OF_STREAM message.
  hr=TransformProcessMessage(pTransform,MFT_MESSAGE_NOTIFY_START_OF_STREAM,
      "MFT_MESSAGE_NOTIFY_START_OF_STREAM",debug);

  if (FAILED(hr))
    goto message;
  /////////////////////////////////////////////////////////////////////////////

  return 0;
// cleanup:
message:
  ResampleChannelDestroy(&pResample->output,"output",debug);
output:
  pBuffer->lpVtbl->Release(pBuffer);
add:
buffer:
  ResampleChannelDestroy(&pResample->input,"input",debug);
input:
  pResamplerProps->lpVtbl->Release(pResamplerProps);
props:
  pTransform->lpVtbl->Release(pTransform);
transform:
  pTransformUnk->lpVtbl->Release(pTransformUnk);
factory:
  return -1;
}

void ResampleDestroy(Resample *pResample)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pResample->pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  IMFTransform *pTransform=pResample->pTransform;

  /////////////////////////////////////////////////////////////////////////////
  TransformProcessMessage(pTransform,MFT_MESSAGE_NOTIFY_END_OF_STREAM,
      "MFT_MESSAGE_NOTIFY_END_OF_STREAM",debug);
  TransformProcessMessage(pTransform,MFT_MESSAGE_COMMAND_DRAIN,
      "MFT_MESSAGE_COMMAND_DRAIN",debug);
  // perform step 7.
  TransformProcessMessage(pTransform,MFT_MESSAGE_NOTIFY_END_STREAMING,
      "MFT_MESSAGE_NOTIFY_END_STREAMING",debug);
  /////////////////////////////////////////////////////////////////////////////

  ResampleChannelDestroy(&pResample->output,"output",debug);
  pResample->pBuffer->lpVtbl->Release(pResample->pBuffer);
  DPUTS(debug,"  input media buffer destroyed\n");
  ResampleChannelDestroy(&pResample->input,"input",debug);

  pResample->pResamplerProps->lpVtbl->Release(pResample->pResamplerProps);
  DPUTS(debug,"  resample properties destroyed\n");
  pResample->pTransform->lpVtbl->Release(pResample->pTransform);
  DPUTS(debug,"  transform destroyed\n");
  pResample->pTransformUnk->lpVtbl->Release(pResample->pTransformUnk);
  DPUTS(debug,"  transform unknown destroyed\n");

  CoTaskMemFree(pResample->pwfx);
}

int ResampleTransform(Resample *pResample)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pResample->pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  Player *pPlayer=pResample->pPlayer;
  IMFTransform *pTransform=pResample->pTransform;
  IMFSample *pSample=pResample->output.pSample;
  IMFMediaBuffer *pBuffer=NULL;
  MFT_OUTPUT_DATA_BUFFER outputDataBuffer;
  DWORD dwStatus;
  DWORD cbCurrentLength;
  BYTE  *pByteBuffer;
  HRESULT hr;
  
DPUTS(debug,"2.1 ...\n");
  for (;;) {
DPUTS(debug,"2.2 ...\n");
    ZeroMemory(
      &outputDataBuffer,
                  // [in]  PVOID Destination,
      sizeof outputDataBuffer
                  // [in]  SIZE_T Length
    );

DPRINTF(debug,"2.3: %p ...\n",pSample);
    outputDataBuffer.pSample=pSample;
    
DPRINTF(debug,"2.4: %p ...\n",pTransform);
    hr=pTransform->lpVtbl->ProcessOutput(pTransform,
      MFT_PROCESS_OUTPUT_DISCARD_WHEN_NO_BUFFER,
                  // [in]       DWORD dwFlags,
      1,          // [in]       DWORD cOutputBufferCount,
      &outputDataBuffer,
                  // [in, out]  MFT_OUTPUT_DATA_BUFFER *pOutputSamples,
      &dwStatus   // [out]      DWORD *pdwStatus
    );

DPUTS(debug,"2.5 ...\n");
    if (MF_E_TRANSFORM_NEED_MORE_INPUT==hr)
      break;
    else if (FAILED(hr)) {
      DERROR(debug,E_UNEXPECTED,hr);
      DERROR(debug,MF_E_INVALIDSTREAMNUMBER,hr);
      DERROR(debug,MF_E_TRANSFORM_STREAM_CHANGE,hr);
      DERROR(debug,MF_E_TRANSFORM_TYPE_NOT_SET,hr);
      DMESSAGE(debug,"processing output");
      goto process;
    }

DPUTS(debug,"2.6 ...\n");
    // convert to a buffer.
    pSample->lpVtbl->ConvertToContiguousBuffer(pSample,
      &pBuffer    // [out]  IMFMediaBuffer **ppBuffer
    );

    if (FAILED(hr)) {
      DERROR(debug,E_UNEXPECTED,hr);
      DMESSAGE(debug,"converting");
      goto convert;
    }

DPUTS(debug,"2.7 ...\n");
    // lock the buffer.
    hr=pBuffer->lpVtbl->Lock(pBuffer,
      &pByteBuffer,
                  // [out]  BYTE **ppbBuffer,
      NULL,       // [out]  DWORD *pcbMaxLength,
      &cbCurrentLength
                  // [out]  DWORD *pcbCurrentLength
    );

    if (FAILED(hr)) {
      DERROR(debug,D3DERR_INVALIDCALL,hr);
      DERROR(debug,MF_E_INVALIDREQUEST,hr);
      DMESSAGE(debug,"locking");
      goto lock;
    }

DPUTS(debug,"2.8 ...\n");
    while (0<=cbCurrentLength) {
      cbCurrentLength-=PlayerWriteWait(pPlayer,(LPSTR)pByteBuffer,
          cbCurrentLength);
    }

DPUTS(debug,"2.9 ...\n");
  // cleanup:
    pBuffer->lpVtbl->Unlock(pBuffer);
  lock:
DPUTS(debug,"2.10 ...\n");
    pBuffer->lpVtbl->Release(pBuffer);
    pBuffer=NULL;
  convert:
  process:
    if (FAILED(hr))
      return -1;
  }

  return 0;
}

int ResampleWrite(Resample *pResample, const void *pData, DWORD cbLength)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pResample->pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  int code=-1;
  IMFMediaBuffer *pBuffer=pResample->pBuffer;
  IMFSample *pSample=pResample->input.pSample;
  IMFTransform *pTransform=pResample->pTransform;
  const char *rp=(const char *)pData;
  const char *mp=rp+cbLength;
  int locked=0;
  DWORD cbMaxLength,cbCurLength;
  BYTE *pByteBuffer;
  HRESULT hr;

  // get the maximum length of the media buffer.
  // always success.
  hr=pBuffer->lpVtbl->GetMaxLength(pBuffer,
    &cbMaxLength                // [out]  DWORD *pcbMaxLength
  );

  // until all data are written ...
  while (rp<mp) {
    // calculate the current length.
    cbCurLength=mp-rp;

    if (cbMaxLength<cbCurLength)
      cbCurLength=cbMaxLength;

//DPRINTF(debug,"cbMaxLength: %d, cbCurLength: %d\n",cbMaxLength,cbCurLength);
DPUTS(debug,"1.1 ...\n");
    // lock the media buffer.
    hr=pBuffer->lpVtbl->Lock(pBuffer,
      &pByteBuffer,             // [out]  BYTE **ppbBuffer,
      NULL,                     // [out]  DWORD *pcbMaxLength,
      NULL                      // [out]  DWORD *pcbCurrentLength
    );

DPUTS(debug,"1.2 ...\n");
    if (FAILED(hr)) {
      DERROR(debug,D3DERR_INVALIDCALL,hr);
      DERROR(debug,MF_E_INVALIDREQUEST ,hr);
      DMESSAGE(debug,"locking media buffer");
      goto lock;
    }

    locked=1;

DPUTS(debug,"1.3 ...\n");
    // copy the current data.
    CopyMemory(
      pByteBuffer,              // _In_  PVOID Destination,
      rp,                       // _In_  const VOID *Source,
      cbCurLength               // _In_  SIZE_T Length
    );

DPUTS(debug,"1.4 ...\n");
#if 1 // {
    // set the current length to the media buffer.
    hr=pBuffer->lpVtbl->SetCurrentLength(pBuffer,
      cbCurLength               // [in]  DWORD cbCurrentLength
    );

    if (FAILED(hr)) {
      DERROR(debug,E_INVALIDARG,hr);
      DMESSAGE(debug,"setting current length to media buffer");
      goto length;
    }
#endif // }

DPUTS(debug,"1.5 ...\n");
    // unlock the media buffer.
    locked=0;
    hr=pBuffer->lpVtbl->Unlock(pBuffer);

    if (FAILED(hr)) {
      DERROR(debug,D3DERR_INVALIDCALL,hr);
      DMESSAGE(debug,"unlocking media buffer");
      goto unlock;
    }

#if 0 // {
    // set the current length to the media buffer.
    hr=pBuffer->lpVtbl->SetCurrentLength(pBuffer,
      cbCurLength               // [in]  DWORD cbCurrentLength
    );

    if (FAILED(hr)) {
      DERROR(debug,E_INVALIDARG,hr);
      DMESSAGE(debug,"setting current length to media buffer");
      goto length;
    }
#endif // }

DPUTS(debug,"1.6 ...\n");
    // setting input stream to transform.
    hr=pTransform->lpVtbl->ProcessInput(pTransform,
      0,                        // [in]  DWORD dwInputStreamID,
      pSample,                  // [in]  IMFSample *pSample,
      0                         // [in]  DWORD dwFlags
    );

    if (FAILED(hr)) {
      DERROR(debug,E_INVALIDARG,hr);
      DERROR(debug,MF_E_INVALIDSTREAMNUMBER,hr);
      DERROR(debug,MF_E_NO_SAMPLE_DURATION,hr);
      DERROR(debug,MF_E_NO_SAMPLE_TIMESTAMP,hr);
      DERROR(debug,MF_E_NOTACCEPTING,hr);
      DERROR(debug,MF_E_TRANSFORM_TYPE_NOT_SET,hr);
      DERROR(debug,MF_E_UNSUPPORTED_D3D_TYPE,hr);
      DMESSAGEV(debug,"processing input (%x)",hr);
      goto process;
    }

DPUTS(debug,"1.7 ...\n");
    if (ResampleTransform(pResample)<0) {
      DMESSAGE(debug,"resampling");
      goto transform;
    }

DPUTS(debug,"1.8 ...\n");
    // increase the read pointer.
    rp+=cbCurLength;
  }

  return cbLength;
// cleanup:
transform:
process:
#if 1 // {
unlock:
length:
#else // } {
length:
unlock:
#endif // }
  if (locked)
    pBuffer->lpVtbl->Unlock(pBuffer);
lock:
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
int ResampleChannelCreate(ResampleChannel *pChannel, WAVEFORMATEX *pwfx,
    ResampleChannelSetType SetType, IMFTransform *pTransform,
    const char *name, int debug)
{
  IMFMediaType *pType;
  IMFSample *pSample;
  HRESULT hr;

  // create the type.
  hr=MediaTypeCreateWFX(&pType,pwfx,name,debug);

  if (FAILED(hr))
    goto create;

  SetType(pTransform,
    0,            // [in]  DWORD dwInputStreamID,
    pType,        // [in]  IMFMediaType *pType,
    0             // [in]  DWORD dwFlags
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_INVALIDMEDIATYPE,hr);
    DERROR(debug,MF_E_INVALIDSTREAMNUMBER,hr);
    DERROR(debug,MF_E_INVALIDTYPE,hr);
    DERROR(debug,MF_E_TRANSFORM_CANNOT_CHANGE_MEDIATYPE_WHILE_PROCESSING,hr);
    DERROR(debug,MF_E_TRANSFORM_TYPE_NOT_SET,hr);
    DERROR(debug,MF_E_UNSUPPORTED_D3D_TYPE,hr);
    DMESSAGEV(debug,"setting % type",name);
    goto set;
  }

  pChannel->pType=pType;
  DPRINTF(debug,"  output %s set\n",name);

  // create the sample.
  hr=MFCreateSample(
    &pSample      // IMFSample **ppIMFSample
  );

  if (FAILED(hr)) {
    DMESSAGEV(debug,"creating %s sample",name);
    goto sample;
  }

  pChannel->pSample=pSample;
  DPRINTF(debug,"  %s sample created\n",name);

  return 0;
// cleanup:
  pSample->lpVtbl->Release(pSample);
sample:
set:
  pType->lpVtbl->Release(pType);
create:
  return -1;
}

void ResampleChannelDestroy(ResampleChannel *pChannel, const char *name,
    int debug)
{
  pChannel->pSample->lpVtbl->Release(pChannel->pSample);
  DPRINTF(debug,"  %s destroyed\n",name);
  pChannel->pType->lpVtbl->Release(pChannel->pType);
  DPRINTF(debug,"  %s type destroyed\n",name);
}

///////////////////////////////////////////////////////////////////////////////
#if 0 // {
void resample_hello()
{
  WAVEFORMATEXTENSIBLE wfxxInput;
  WAVEFORMATEXTENSIBLE wfxxOutput;
  Resample res;
  HRESULT hr;

  // starting up media foundation.
  hr=MFStartup(
    MF_VERSION_VISTA,               // ULONG Version,
    MFSTARTUP_NOSOCKET              // DWORD dwFlags = MFSTARTUP_FULL
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_BAD_STARTUP_VERSION,hr);
    DERROR(debug,MF_E_DISABLED_IN_SAFEMODE,hr);
    DERROR(debug,E_NOTIMPL,hr);
    DMESSAGE(debug,"starting up media foundation");
    goto startup;
  }

  DPUTS(debug,"  media foundation started up\n");

  // set attributes.
  WaveFormatExtensibleSetup(&wfxxInput,44100,2,24);
  WaveFormatExtensibleSetup(&wfxxOutput,48000,2,24);

  // create the resampler.
  if (ResampleCreate(&res,&wfxxInput.Format,&wfxxOutput.Format)<0)
    goto res;

// cleanup:
  ResampleDestroy(&res);
res:
  MFShutdown();
  DPUTS(debug,"  media foundation shut down\n");
startup:
  ;
}
#endif // }
#elif defined (YASAPI_DEBUG) // {
void resample_hello()
{
}
#endif // }
